//Crayonizer. A coloration/formatting app for command-line output
//Written by Colum Paget.
//Copyright 2013 Colum Paget.

/****  Gnu Public Licence ****/
/*
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "common.h"
#include "config_file.h"
#include "help.h"
#include <sys/ioctl.h>

#define NORM "\x1b[0m"
#define CLRSCR "\x1b[2J\x1b[;H"


char *Version="0.0.6";
int GlobalFlags=0;
TCrayon *KeyPresses=NULL;
int NoOfKeyPresses=0;
ListNode *Vars=NULL;
int StartTime=0;



void HandleSigwinch(STREAM *Pipe)
{
    struct winsize w;
		char *Tempstr=NULL;

    ioctl(0, TIOCGWINSZ, &w);

		//if we don't do this, then vim causes trouble. I don't understand why.
		w.ws_row++;	
    ioctl(Pipe->out_fd, TIOCSWINSZ, &w);


		DestroyString(Tempstr);
}


void HandleSignal(int sig)
{
int wid, len;

if (sig==SIGWINCH) GlobalFlags |= GOT_SIGWINCH;
if (sig==SIGTERM) GlobalFlags |= GOT_SIGTERM;
if (sig==SIGINT) GlobalFlags |= GOT_SIGINT;
}

void PropogateSignals(STREAM *Pipe)
{
int PeerPID;

PeerPID=atoi(STREAMGetValue(Pipe,"PeerPID"));
if (GlobalFlags & GOT_SIGWINCH) 
{
HandleSigwinch(Pipe);
//kill(PeerPID,SIGWINCH);
}

if (GlobalFlags & GOT_SIGTERM) kill(PeerPID,SIGTERM);
if (GlobalFlags & GOT_SIGINT) kill(PeerPID,SIGINT);
GlobalFlags &= ~(GOT_SIGWINCH | GOT_SIGTERM | GOT_SIGINT);
}


char *VtCode(int Color, int BgColor, int Flags)
{
static char *ANSI=NULL;

if ((! Color) && (! Flags)) return("");

if ( (Color > 0) && (BgColor > 0) )
{	
	BgColor=BgColor >> 8;
	ANSI=FormatStr(ANSI,"\x1b[%d;%d",30+Color-1,40+BgColor-1);
	if (Flags) ANSI=CatStr(ANSI,";");
}
else if (Color > 0) 
{
	ANSI=FormatStr(ANSI,"\x1b[%d",30+Color-1);
	if (Flags) ANSI=CatStr(ANSI,";");
}
else ANSI=CopyStr(ANSI,"\x1b[");

if (Flags)
{
	if (Flags & FLAG_BOLD) ANSI=CatStr(ANSI,"01");
	if (Flags & FLAG_FAINT) ANSI=CatStr(ANSI,"02");
	if (Flags & FLAG_UNDER) ANSI=CatStr(ANSI,"04");
	if (Flags & FLAG_BLINK) ANSI=CatStr(ANSI,"05");
	if (Flags & FLAG_INVERSE) ANSI=CatStr(ANSI,"07");
}
ANSI=CatStr(ANSI,"m");

return(ANSI);
}


char *RebuildPath(char *RetStr, char *Path, char *CurrDir)
{
char *NewPath=NULL, *Token=NULL, *ptr;

NewPath=CopyStr(RetStr,"");
ptr=GetToken(Path,":",&Token,0);
while (ptr)
{
	if (strcmp(Token,CurrDir) !=0) NewPath=MCatStr(NewPath,Token,":",NULL);
	ptr=GetToken(ptr,":",&Token,0);
}

DestroyString(Token);
return(NewPath);
}


char *RebuildCommandLine(char *RetStr, int argc, char *argv[])
{
char *CommandLine=NULL, *Quoted=NULL;
int i;


CommandLine=MCopyStr(RetStr, argv[0]," ",GetVar(Vars,"ExtraCmdLineOptions")," ",NULL);
for (i=1; i < argc; i++)
{
	Quoted=QuoteCharsInStr(Quoted,argv[i]," 	&|");
	CommandLine=MCatStr(CommandLine,Quoted," ",NULL);
}

return(CommandLine);
return(Quoted);
}


void LoadEnvironment()
{
char *Token=NULL, *ptr;
int i;

Vars=ListCreate();
for (i=0; environ[i] !=NULL; i++)
{
ptr=GetToken(environ[i],"=",&Token,0);
SetVar(Vars,Token,ptr);
}

DestroyString(Token);
}


int HandleCrayonIf(TCrayon *Crayon)
{
char *Expr=NULL, *Tempstr=NULL, *Token=NULL, *PrevToken=NULL, *ptr, *tptr;
int result=FALSE, val;

	Expr=SubstituteVarsInString(Expr,Crayon->Match,Vars,0);

	ptr=GetToken(Expr,"\\S",&Token,0);
	while (ptr)
	{
		switch (Token[0])
		{
			case 'e':
				if (strncmp(Token,"exists(",7)==0)
				{
					Tempstr=CopyStr(Tempstr,Token+7);
					tptr=strrchr(Tempstr,')');
					if (tptr) *tptr='\0';

					if (access(Tempstr,F_OK)==0) result=TRUE;
					else result=FALSE;
				}
			break;

			case 'i':
				if (strncmp(Token,"isatty(",7)==0)
				{
					Tempstr=CopyStr(Tempstr,Token+7);
					tptr=strrchr(Tempstr,')');
					if (tptr) *tptr='\0';

					if (strcmp(Tempstr,"stdin")==0) val=0;
					else val=1;
					result=isatty(val);
				}
			break;

			case 'n':
				if (strncmp(Token,"notatty(",8)==0)
				{
					Tempstr=CopyStr(Tempstr,Token+8);
					tptr=strrchr(Tempstr,')');
					if (tptr) *tptr='\0';
					if (strcmp(Tempstr,"stdin")==0) val=0;
					else val=1;
					result= !isatty(val);
				}
			break;


			case '<':
				ptr=GetToken(ptr,"\\S",&Token,0);
				result=(atoi(PrevToken) < atoi(Token));
			break;

			case '>':
				ptr=GetToken(ptr,"\\S",&Token,0);
				result=(atoi(PrevToken) < atoi(Token));
			break;


			case '=':
				ptr=GetToken(ptr,"\\S",&Token,0);
				if (strcasecmp(PrevToken, Token)==0) result=TRUE;
				else result=FALSE;
			break;


			case '!':
				ptr=GetToken(ptr,"\\S",&Token,0);
				if (strcasecmp(PrevToken, Token)==0) result=TRUE;
				else result=FALSE;
			break;

		}
		PrevToken=CopyStr(PrevToken,Token);
		ptr=GetToken(ptr,"\\S",&Token,0);
	}


DestroyString(Expr);
DestroyString(Token);
DestroyString(PrevToken);
DestroyString(Tempstr);
return(result);
}

int CrayonMatches(TCrayon *Crayon, char *sptr, char *eptr)
{
char *ptr;
int i, result=FALSE;
float val;


switch (Crayon->Type)
{
//typedef enum {CRAYON_ACTION, CRAYON_ARGS, CRAYON_LINE, CRAYON_STRING, CRAYON_VALUE, CRAYON_MAPTO, CRAYON_LINEMAPTO, CRAYON_APPEND, CRAYON_PREPEND, CRAYON_EXEC, CRAYON_CMDLINE, CRAYON_PASSINPUT, CONFIG_INCLUDE, CRAYON_IF};
//These always 'match', as they are not checks on the value of anything

case CRAYON_ACTION:
case CRAYON_ARGS:
case CRAYON_LINE:
case CRAYON_STRING:
case CRAYON_MAPTO:
case CRAYON_LINEMAPTO:
case CRAYON_APPEND: 
case CRAYON_PREPEND: 
case CRAYON_PASSINPUT: 
case CRAYON_EXEC: 
case CRAYON_CMDLINE: 
case CRAYON_VALUE: 
	return(TRUE);
break;


case CRAYON_COMPARATOR:
	val=atof(sptr);
//For a 'value' item, it's actions can contain multiple 'value' checks
	switch (Crayon->Op)
	{
	case '<':
	if (val < Crayon->Value) result=TRUE;
	break;
	
	case '>':
	if (val > Crayon->Value) result=TRUE;
	
	break;
	
	case '=':
	if (val == Crayon->Value) result=TRUE;
	break;
	
	case '!':
	if (val != Crayon->Value) result=TRUE;
	break;
	}
break;


case CRAYON_IF:
	result=HandleCrayonIf(Crayon);
break;

}

return(result);
}


void PassToProgram(STREAM *Pipe, char *ActivateLine, char *Program)
{
STREAM *S, *CmdS, *CmdErr;
char *Tempstr=NULL;
int result, wrote;
ListNode *Streams;


Streams=ListCreate();
CmdS=STREAMCreate();
CmdErr=STREAMCreate();
PipeSpawn( &(CmdS->out_fd), &(CmdS->in_fd), &(CmdErr->in_fd), Program);

ListAddItem(Streams,Pipe);
ListAddItem(Streams,CmdS);
ListAddItem(Streams,CmdErr);
Tempstr=SetStrLen(Tempstr,4096);
STREAMFlush(Pipe);

write(CmdS->out_fd,ActivateLine,StrLen(ActivateLine));
while (1)
{
	S=STREAMSelect(Streams);
	
	if (S)
	{
		if (S==CmdS) 
		{
			result=STREAMReadBytes(CmdS,Tempstr,4096);
			if (result < 1) break;
			write(Pipe->out_fd,Tempstr,result);
		}

		if (S==CmdErr) 
		{
			result=STREAMReadBytes(CmdErr,Tempstr,4096);
			if (result < 1) break;
			write(1,Tempstr,result);
		}

		if (S==Pipe)
		{
			result=STREAMReadBytes(Pipe,Tempstr,4096);
			wrote=write(CmdS->out_fd,Tempstr,result);
			if (wrote < result) break;
		}
	}
}

STREAMClose(S);
DestroyString(Tempstr);
}


void ApplySingleAction(STREAM *Pipe, int *AttribLine, char *Line, char *MatchStart, char *MatchEnd, TCrayon *Action)
{
int Attribs;
int start, end, i;
char *Tempstr=NULL, *EnvName=NULL, *ptr;

if (GlobalFlags & FLAG_DONTCRAYON) return;

	Attribs=Action->Attribs;
	start=MatchStart-Line;
	end=MatchEnd-Line;


		//These have to be done when we have the particular match in hand, they 
		//can't be done when we output the line at the end


		switch (Action->ActType)
		{
		case ACTION_BASENAME:
			ptr=strrchr(MatchStart,'/');
			if (ptr && (ptr < MatchEnd))
			{
				MatchStart=ptr+1;
				start=MatchStart-Line;
			}
		break;

		case ACTION_SET_XTITLE: 
			Tempstr=CopyStr(Tempstr,"\x1b]0;");
			Tempstr=CatStrLen(Tempstr,MatchStart,end-start);
			StripTrailingWhitespace(Tempstr);
			Tempstr=CatStr(Tempstr,"\x07");
			write(1,Tempstr,StrLen(Tempstr));
		break;

		case ACTION_RESTORE_XTITLE:
			Tempstr=CopyStr(Tempstr,"\x1b]0;");
			Tempstr=CatStr(Tempstr,GetVar(Vars,"OldXtermTitle"));
			StripTrailingWhitespace(Tempstr);
			Tempstr=CatStr(Tempstr,"\x07");
			write(1,Tempstr,StrLen(Tempstr));
		break;

		case ACTION_SETENV:
			ptr=strchr(Action->String,'=');
			if (ptr) 
			{
				EnvName=CopyStrLen(EnvName,Action->String,ptr-Action->String);
				Tempstr=CopyStr(Tempstr,ptr+1);
			}
			else
			{
				EnvName=CopyStr(EnvName,Action->String);
				Tempstr=CopyStrLen(Tempstr,MatchStart,end-start);
			}
			StripTrailingWhitespace(Tempstr);
			setenv(EnvName,Tempstr,TRUE);
			SetVar(Vars,EnvName,Tempstr);
		break;

		case ACTION_ARGS:
			Tempstr=SubstituteVarsInString(Tempstr,Action->String,Vars,0);
			SetVar(Vars,"ExtraCmdLineOptions",Tempstr);
		break;

		case ACTION_PLAYSOUND:
			Tempstr=SubstituteVarsInString(Tempstr,Action->String,Vars,0);
			SoundPlayFile("/dev/dsp",Action->String,VOLUME_LEAVEALONE, PLAYSOUND_NONBLOCK);
		break;

		case ACTION_EXEC:
			Tempstr=SubstituteVarsInString(Tempstr,Action->String,Vars,0);
		break;

		case ACTION_PASSTO:
			Tempstr=SubstituteVarsInString(Tempstr,Action->String,Vars,0);
			PassToProgram(Pipe, Line, Tempstr);
		break;

		case ACTION_SEND:
			Tempstr=SubstituteVarsInString(Tempstr,Action->String,Vars,0);
			STREAMWriteLine(Tempstr,Pipe); STREAMFlush(Pipe);
			usleep(250000);
		break;

		case ACTION_ECHO:
			Tempstr=SubstituteVarsInString(Tempstr,Action->String,Vars,0);
			write(1,Tempstr,StrLen(Tempstr));
		break;


		case ACTION_ALTSCREEN:
			write(1,"\x1b[?47h",6); 
		break;

		case ACTION_NORMSCREEN:
			write(1,"\x1b[?47l",6); 
		break;

		case ACTION_XTERM_FGCOLOR:
			EnvName=SubstituteVarsInString(EnvName,Action->String,Vars,0);
			Tempstr=MCopyStr(Tempstr,"\x1b]10;",EnvName,"\007",NULL);
			write(1,Tempstr,StrLen(Tempstr));
		break;


		case ACTION_XTERM_BGCOLOR:
			EnvName=SubstituteVarsInString(EnvName,Action->String,Vars,0);
			Tempstr=MCopyStr(Tempstr,"\x1b]11;",EnvName,"\007",NULL);
			write(1,Tempstr,StrLen(Tempstr));
		break;

		case ACTION_RXVT_FGCOLOR:
			EnvName=SubstituteVarsInString(EnvName,Action->String,Vars,0);
			Tempstr=MCopyStr(Tempstr,"\x1b]39;",EnvName,"\007",NULL);
			write(1,Tempstr,StrLen(Tempstr));
		break;

		case ACTION_RXVT_BGCOLOR:
			EnvName=SubstituteVarsInString(EnvName,Action->String,Vars,0);
			Tempstr=MCopyStr(Tempstr,"\x1b]49;",EnvName,"\007",NULL);
			write(1,Tempstr,StrLen(Tempstr));
		break;


		case ACTION_CLEARSCREEN:
			write(1,CLRSCR,StrLen(CLRSCR)); 
		break;


		case ACTION_DONTCRAYON:
			//use '=' not '|=' here because we don't want it to honor
			//any other flags
			GlobalFlags = FLAG_DONTCRAYON;
		break;
		}


		if (AttribLine) for (i=start; i < end; i++) AttribLine[i] |= Attribs;

	DestroyString(Tempstr);
	DestroyString(EnvName);
}


void ApplyActions(STREAM *Pipe, int *AttribLine, char *Line, char *MatchStart, char *MatchEnd, TCrayon *Crayon)
{
int i,sum=0;
char *ptr;

if (GlobalFlags & FLAG_DONTCRAYON) return;


	//for map to we must calculate which attribute to use
	if ((Crayon->Type==CRAYON_MAPTO) || (Crayon->Type==CRAYON_LINEMAPTO))
	{
		for (ptr=MatchStart; ptr < MatchEnd; ptr++) sum+=*ptr;
		ApplySingleAction(Pipe, AttribLine, Line, MatchStart, MatchEnd, &Crayon->Actions[sum % Crayon->ActionCount]);
	}
	else 
	{
		if (CrayonMatches(Crayon, MatchStart,MatchEnd)) 
		{
			ApplySingleAction(Pipe,AttribLine, Line, MatchStart, MatchEnd, Crayon);
			for (i=0; i < Crayon->ActionCount; i++)
			{
				ApplyActions(Pipe, AttribLine, Line, MatchStart, MatchEnd, &Crayon->Actions[i]);
			}
		}
	}

}


//AttribLine is the string of attribute values for each char in Line
void ColorSubstring(STREAM *Pipe, int *AttribLine, char *Line, int Len, TCrayon *Crayon)
{
char *Tempstr=NULL, *p_LineEnd;
int result;
ListNode *Matches, *Curr;
TPMatch *Match;
int Attribs;

p_LineEnd=Line+Len;

Matches=ListCreate();
result=pmatch(Crayon->Match, Line, p_LineEnd-Line, Matches, PMATCH_SUBSTR);
if (result > 0)
{
	Curr=ListGetNext(Matches);
	while (Curr)
	{
		Match=(TPMatch *) Curr->Item;
		ApplyActions(Pipe, AttribLine, Line, Match->Start, Match->End, Crayon);
	Curr=ListGetNext(Curr);
	}
}

ListDestroy(Matches,DestroyString);
}


void OutputLineWithAttributes(char *Line, int *Attribs, int Len)
{
int i, LastAttrib=0;
char *Tempstr=NULL;

if (! Attribs) return;
if (Attribs[0] & FLAG_HIDE) return;

for (i=0; i < Len; i++)
{
	if (Attribs[i] != LastAttrib) 
	{
		Tempstr=CatStr(Tempstr,NORM);
		if (Attribs[i] > 0) Tempstr=CatStr(Tempstr,VtCode(Attribs[i] & 0x00FF, Attribs[i] & 0xFF00, Attribs[i] & 0xFF0000));
		LastAttrib=Attribs[i];
	}
	if (LastAttrib & FLAG_CAPS) Tempstr=AddCharToStr(Tempstr,toupper(Line[i]));
	else if (LastAttrib & FLAG_LOWERCASE) Tempstr=AddCharToStr(Tempstr,tolower(Line[i]));
	else Tempstr=AddCharToStr(Tempstr,Line[i]);
}

Tempstr=CatStr(Tempstr,NORM);
write(1,Tempstr,StrLen(Tempstr));

DestroyString(Tempstr);
}



void ColorLine(STREAM *Pipe, char *Line, int Len, ListNode *ColorMatches)
{
ListNode *Curr;
int *Attribs=NULL;
char *p_LineEnd;
int i, LastAttrib=0;
TCrayon *Crayon;

Attribs=(int *) calloc(Len,sizeof(int));
Curr=ListGetNext(ColorMatches);
while (Curr)
{
	Crayon=(TCrayon *) Curr->Item;
	switch (Crayon->Type)
	{
	case CRAYON_LINE:
	case CRAYON_LINEMAPTO:
	if (pmatch(Curr->Tag,Line,Len,NULL,PMATCH_SUBSTR) > 0) ApplyActions(Pipe, Attribs, Line, Line, Line+Len, Crayon);
	break;

	case CRAYON_VALUE:
	case CRAYON_STRING:
	case CRAYON_MAPTO:
		ColorSubstring(Pipe, Attribs, Line, Len, Crayon);
	break;
	}


	Curr=ListGetNext(Curr);
}

OutputLineWithAttributes(Line, Attribs, Len);

free(Attribs);
}


int HandleStdin(STREAM *StdIn, STREAM *Out)
{
int result, i;
char *Tempstr=NULL;

Tempstr=SetStrLen(Tempstr,4096);
result=STREAMReadBytes(StdIn,Tempstr,4096);
if (result > 0) 
{
	for (i=0; i < NoOfKeyPresses; i++)
	{
		if (strcmp(KeyPresses[i].Match,Tempstr)==0) 
		{
			ApplyActions(Out,NULL, NULL, NULL, NULL, KeyPresses+i); 
			break;
		}
	}

	if (i==NoOfKeyPresses)
	{
	STREAMWriteBytes(Out,Tempstr,result);
	STREAMFlush(Out);
	}
}

DestroyString(Tempstr);
return(result);
}


int ColorProgramOutput(STREAM *Pipe, ListNode *ColorMatches)
{
	char *Tempstr=NULL, *sptr, *eptr;
	int result, len;

	Tempstr=SetStrLen(Tempstr,4098);
	result=STREAMReadBytes(Pipe,Tempstr,4096);
	if (result > 0)
	{
	Tempstr[result]='\0';

	sptr=Tempstr;
	while (sptr && (*sptr != '\0'))
	{
		eptr=strchr(sptr,'\n');
		if (! eptr) eptr=Tempstr+result;
		else eptr++;
		len=eptr-sptr;
		if (len >0) ColorLine(Pipe, sptr,len,ColorMatches);
		sptr=eptr;
	}
	}

	DestroyString(Tempstr);
	return(result);
}



void ProcessAppends(STREAM *Pipe, ListNode *Crayons, int Type)
{
TCrayon *Item;
ListNode *Curr;
char *Tempstr=NULL;
int *Attribs=NULL;
int i, Len;

Curr=ListGetNext(Crayons);
while (Curr)
{
	Item=(TCrayon *) Curr->Item;

	if (
			(Item->Type==Type) ||
			( (Type==CRAYON_PREPEND) && ((Item->Type==CRAYON_IF) || (Item->Type==CRAYON_ARGS)) )
		)
	{
		Tempstr=SubstituteVarsInString(Tempstr,Item->Match,Vars,0);
		Tempstr=CatStr(Tempstr,"\n");
		Len=StrLen(Tempstr);
		Attribs=realloc(Attribs,Len * sizeof(int));
		memset(Attribs,0,Len*sizeof(int));

		ApplyActions(Pipe, Attribs, Tempstr, Tempstr, Tempstr+Len,Item); 
		if ((Item->Type==CRAYON_PREPEND) || (Item->Type==CRAYON_APPEND)) OutputLineWithAttributes(Tempstr, Attribs, Len);

	}
Curr=ListGetNext(Curr);
}

DestroyString(Tempstr);
if (Attribs) free(Attribs);
}	



void ProcessCmdLine(char *CmdLine, ListNode *Crayons)
{
TCrayon *Item;
ListNode *Curr;

Curr=ListGetNext(Crayons);
while (Curr)
{
	Item=(TCrayon *) Curr->Item;
	if (Item->Type==CRAYON_CMDLINE)
	{
		 ColorSubstring(NULL, NULL, CmdLine, StrLen(CmdLine), Item);
	}
	Curr=ListGetNext(Curr);
}

}	





STREAM *LaunchCommands(int argc, char *argv[], ListNode *Matches)
{
char *Tempstr=NULL;
ListNode *Curr;
STREAM *Pipe;
TCrayon *Crayon;
int i;

Curr=ListGetNext(Matches);
while (Curr)
{
Crayon=(TCrayon *) Curr->Item;

if (Crayon->Type==CRAYON_EXEC) 
{
	if (StrLen(Tempstr)) Tempstr=MCatStr(Tempstr,";",Crayon->Match,NULL);
	else Tempstr=CopyStr(Tempstr,Crayon->Match);
}

Curr=ListGetNext(Curr);
}


Tempstr=RebuildCommandLine(Tempstr,argc, argv);

if (GlobalFlags & FLAG_DONTCRAYON)
{
	//Exit after running the command
	exit(system(Tempstr));
}

Pipe=STREAMSpawnCommand(Tempstr,COMMS_BY_PTY|SPAWN_TRUST_COMMAND|TTYFLAG_ECHO|TTYFLAG_CANON|TTYFLAG_CRLF);
signal(SIGTERM, HandleSignal);
signal(SIGINT, HandleSignal);
signal(SIGWINCH, HandleSignal);

//Set initial window size, as though we'd received a SIGWINCH
HandleSigwinch(Pipe);

DestroyString(Tempstr);
return(Pipe);
}


//Currently only reads Title
void XTermReadValue()
{
char *Tempstr=NULL, *ptr;
STREAM *StdIn;

StdIn=STREAMFromFD(0);
STREAMSetTimeout(StdIn,1);
Tempstr=CopyStr(Tempstr,"\x1b[21;t");
write(1,Tempstr,StrLen(Tempstr));
Tempstr=STREAMReadToTerminator(Tempstr,StdIn,'\\');
if (StrLen(Tempstr))
{
	ptr=Tempstr+StrLen(Tempstr)-2;
	if (strcmp(ptr,"\x1b\\")==0) *ptr='\0';
	ptr=Tempstr;
	if (strncmp(ptr,"\x1b]l",3)==0) ptr+=3;
	if (StrLen(ptr)) SetVar(Vars,"OldXtermTitle",ptr);
}

STREAMDisassociateFromFD(StdIn);
DestroyString(Tempstr);
}


void LoadConfig(char *CmdLine, char **CrayonizerDir, ListNode *ColorMatches)
{
int i;
char *Tempstr=NULL;

Tempstr=MCopyStr(Tempstr,GetCurrUserHomeDir(),"/",USER_CONFIG_FILE,NULL);
if (! ConfigReadFile(Tempstr, CmdLine, CrayonizerDir, ColorMatches))
{
	if (! ConfigReadFile(GLOBAL_CONFIG_FILE, CmdLine, CrayonizerDir, ColorMatches))
	{
		printf("ERROR! Crayonizer can't find config file. Tried %s and %s\n",Tempstr, GLOBAL_CONFIG_FILE);
		exit(1);
	}
}

}



//Spawns a command, reads text from it and 'Crayonizes' it
void CrayonizeCommand(int argc, char *argv[])
{
STREAM *StdIn=NULL, *Pipe=NULL, *S;
char *Tempstr=NULL, *CrayonizerDir=NULL, *CmdLine=NULL;
ListNode *ColorMatches, *Streams;
int val, i, result;

ColorMatches=ListCreate();
for (i=0; i < argc; i++) CmdLine=MCatStr(CmdLine,argv[i]," ",NULL);
StripTrailingWhitespace(CmdLine);

LoadConfig(CmdLine, &CrayonizerDir, ColorMatches);

if (! StrLen(CrayonizerDir))
{
		printf("ERROR! You must specify 'CrayonizerDir' in Crayonizer Config File\n");
		printf("Failure to do so can result in crayonizer calling itself resulting in\n");
		printf("an infinite loop of crayonizer processes (forkbomb)\n");
		printf("AND YOU DON'T WANT THAT!\n");
		exit(1);
}

Tempstr=RebuildPath(Tempstr,getenv("PATH"),CrayonizerDir);
setenv("PATH",Tempstr,TRUE);
LoadEnvironment();

//if not outputing to a tty then run
if (! isatty(1)) GlobalFlags |=FLAG_DONTCRAYON;
if (! ListSize(ColorMatches)) GlobalFlags |=FLAG_DONTCRAYON;


//Okay, we're committed to running the command!
Streams=ListCreate();


//if we are going to restore the title at the end, then we need to set it here
//if (GlobalFlags & FLAG_RESTORE_XTITLE) XTermReadValue();


ProcessCmdLine(CmdLine, ColorMatches);
ProcessAppends(Pipe, ColorMatches,CRAYON_PREPEND);


Pipe=LaunchCommands(argc, argv, ColorMatches);


if (GlobalFlags & FLAG_PASSINPUT)
{
	InitTTY(0, 0,TTYFLAG_LFCR|TTYFLAG_IGNSIG);
	StdIn=STREAMFromFD(0);
	ListAddItem(Streams,StdIn);
}


ListAddItem(Streams,Pipe);

while (1)
{
	S=STREAMSelect(Streams);

	if (S)
	{
		if (S==StdIn) result=HandleStdin(StdIn,Pipe);
		else result=ColorProgramOutput(Pipe, ColorMatches);
		if (result < 0) break;
	}
	PropogateSignals(Pipe);
}
wait(&val);

Tempstr=FormatStr(Tempstr,"%d",time(NULL) - StartTime);
SetVar(Vars,"duration",Tempstr);
ProcessAppends(Pipe,ColorMatches,CRAYON_APPEND);

STREAMClose(Pipe);
STREAMDisassociateFromFD(StdIn);
ResetTTY(0);
ListDestroy(Streams,NULL);
ListDestroy(ColorMatches,free);
DestroyString(Tempstr);
DestroyString(CmdLine);
DestroyString(CrayonizerDir);
}



void CrayonizeSTDIN(int argc, char *argv[])
{
STREAM *StdIn=NULL;
char *Tempstr=NULL, *CrayonizerDir=NULL, *CmdLine=NULL;
ListNode *ColorMatches;
int val, i;

ColorMatches=ListCreate();
for (i=0; i < argc; i++) CmdLine=MCatStr(CmdLine,argv[i]," ",NULL);
LoadConfig(CmdLine, &CrayonizerDir, ColorMatches);

StdIn=STREAMFromFD(0);

Tempstr=STREAMReadLine(Tempstr,StdIn);
while (Tempstr)
{
ColorLine(StdIn, Tempstr,StrLen(Tempstr),ColorMatches);
Tempstr=STREAMReadLine(Tempstr,StdIn);
}

STREAMDisassociateFromFD(StdIn);
ListDestroy(ColorMatches,free);
DestroyString(CrayonizerDir);
DestroyString(CmdLine);
DestroyString(Tempstr);
}



void CalledAsSelf(int argc, char *argv[])
{
char *Args[]={"-v","-version","--version","-?","-h","-help","--help","-config-help","-pmatch-help","-stdin",NULL};
typedef enum {ARG_V1, ARG_V2, ARG_V3, ARG_H1, ARG_H2, ARG_H3, ARG_H4, ARG_CONFIG_HELP, ARG_PMATCH_HELP, ARG_STDIN};
int i, val, new_argc=0;
char **new_argv=NULL;


for (i=1; i < argc; i++)
{
	val=MatchTokenFromList(argv[i],Args,0);
	
	switch(val)
	{
		case ARG_V1:
		case ARG_V2:
		case ARG_V3:
			printf("crayonizer (Crayonizer) %s\n",Version);
		break;
		
		case ARG_H1:
		case ARG_H2:
		case ARG_H3:
		case ARG_H4:
			PrintUsage();
		break;

		case ARG_CONFIG_HELP:
			PrintConfigFileHelp();
		break;

		case ARG_PMATCH_HELP:
			PrintPMatchHelp();
		break;

		case ARG_STDIN:
			CrayonizeSTDIN(argc, argv);
		break;

		default:
			new_argc++;
			new_argv=(char **) realloc(new_argv,sizeof(char *) * new_argc);
			new_argv[new_argc-1]=argv[i];
		break;
	}
	
}

if (new_argc > 0) CrayonizeCommand(new_argc,new_argv);
}



main(int argc, char *argv[])
{
char *ptr;

time(&StartTime);

if (strcmp(basename(argv[0]),"crayonizer")==0) CalledAsSelf(argc, argv);
else 
{
	ptr=basename(argv[0]);
	argv[0]=ptr;
	CrayonizeCommand(argc,argv);
}
}
